/*
 * beam.c
 * code to beam texts across the TCP connection following a special protocol
 * source file for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>

#include "cancan.h"
#include "term.h"
#include "beam.h"

int pause();
void bzero();
#ifdef NOBZERO
#define bzero(a,b) memset ((a), 0, (b))
#endif

static int next_num();
static void edit_message();
static void finish_edit();

editsess *edit_sess;		/* head of session list */

char *edit_start, *edit_end;	/* when starting/leaving editing sessions */ 

/*
 * read in another decimal. return 0 in case of overflow or invalid char.
 */
static int next_num(val, cp)
unsigned long *val;
char *cp;
{
    unsigned int i;
    return (i = ctoi(*cp)) < 10 ? (*val = *val * 10 + i, 1) : 0;
}

/*
 * Process editing protocol message with len chars.
 * Return number of characters used in the message.
 */
int mp_receive(sesp, data, len)
editsess **sesp;
char *data;
int len;
{
    static char mbuf[64];
    editsess *s = *sesp;
    char *src = data;
    int state, l;
    if(!s) {
	*sesp = s = (editsess*)malloc(sizeof(editsess));
	if(!s) syserr("malloc");
	bzero((char *)s, sizeof(editsess));
	if(edit_sess) {
	    s->num = edit_sess->num;
	    s->next = edit_sess;
	}
	s->num++;  edit_sess = s;
	src += strlen(MPI);  len -= strlen(MPI);
    }
    if((state = s->state) >= 2 && len > (l = s->size - s->pos)) len = l;
    l = len;
    while(l) {
	switch(state) {
	case 0:					/* 'E' or 'V'	*/
		if(*src != 'E' && *src != 'V') break;
		s->mode = *src++;  ++state;  if(!--l) break;
		/* FALLTHROUGH */
	case 1:					/* message size	*/
		while(next_num(&s->size, src) && (++src, --l));
		if(!l || !s->size) break;  ++state;
		if(l > s->size) l = s->size;
		continue;
	case 4:					/* 'M' if edit	*/
		len = l;	/* message body starts */
		if(s->mode != 'E') { state = 8;  continue; }
		if(*src != 'M') break;
		++src;  ++state;  if(!--l) break;
		/* FALLTHROUGH */
	case 5:					/* message key	*/
		while(next_num(&s->key, src) && (++src, --l));
		if(!l || !s->key) break;  ++state;  continue;
	case 8:					/* description	*/
		for( ; isprint(*src) && l; ++src, --l)
		    if(s->dlen < MAXDESCLEN) s->desc[s->dlen++] = *src;
		if(!l) break;
		printf("%.*s", s->dlen, s->desc);  newline();
		++state;
		/* FALLTHROUGH */
	case 2: case 6: case 9:			/* cr's		*/
		++state;  if(*src == '\r' && (++src, !--l)) break;
		/* FALLTHROUGH */
	case 3: case 7: case 10:		/* lf's		*/
		if(*src != '\n') break;
		++state;  ++src;  --l;
		if(state != 11) continue;
		/* FALLTHROUGH */
	case 11:				/* create file	*/
		sprintf(mbuf, "/tmp/%c%lu", s->mode,
			s->key ? s->key : (unsigned long)time(NULL));
		if((s->fd = open(mbuf, O_WRONLY | O_CREAT, 0666)) >= 0)
		    { s->fname = str_dup(mbuf);  ++state;  continue; }
		printf("#could not create temp file\n");
		s->mode = 'C';  break;
	case 12:				/* write file	*/
		if(s->mode == 'C') {
		    if(len < 256) break; /* assume short packet -> msg ends */
		} else if(write(s->fd, src, l) < l) {
		    printf("#could not write temp file\n");
		    s->mode = 'C';
		}
		src += l;  l = 0;
		if(s->pos + len < s->size) {
		    sprintf(footnote, "read %lu bytes out of %lu",
				s->pos + len, s->size);
		}
	} /* end switch */
	break;
    } /* end while */
    if(((s->state = state) > 4 && (s->pos += len - l) >= s->size) || l) {
	*sesp = 0;
	if(l || state < 11) {
	    if(s->mode != 'C')
		strcpy(footnote, "#warning: mume protocol error");
	    finish_edit(s);
	} else
	    edit_message(s);
    }
    return src - data;
}    

static void edit_message(s)
editsess *s;
{
    char *editor = 0;
    int i, waitforeditor = 0;
    footnote[0] = 0;
    if(s->mode == 'E') {
	struct stat sbuf;
	if(fstat(s->fd, &sbuf) < 0) syserr("fstat");	/* file is open */
	s->size = sbuf.st_size;
	s->time = sbuf.st_mtime;
	if(!(editor = getenv("EDITOR"))) editor = DEFAULT_EDITOR;
    } else if(s->mode == 'V') {
	if(!(editor = getenv("PAGER"))) editor = DEFAULT_PAGER;
    }
    close(s->fd);
    if(!editor) {
	finish_edit(s);
	return;
    }
    /* send an edit_start message (if wanted) */ 
    if(!edit_sess->next && edit_start)
	parse(edit_start, 0, NULL);
    if(editor[0] == '&')
	editor++;
    else {
	waitforeditor = 1;
	/* ignore SIGINT lest child interrupt kill us too */
	signal(SIGINT, SIG_IGN);
	reset_terminal();
    }
    fflush(stdout);
    waiting_for_pid = fork();
    if(!waiting_for_pid) {
	char *args[MAXARGS + 1];
	char *p = strtok(editor, " ");
	for(i = 0; p && i < MAXARGS - 1; p = strtok(NULL, " "), i++)
	    args[i] = p;
	args[i] = s->fname;  args[i + 1] = 0;
	if(!waitforeditor) { close(0);  close(1); }
	execvp(args[0], args);
	printf("#could not execute \"%s\"\n", args[0]);  fflush(stdout);
	_exit(1);
    } else if(waiting_for_pid < 0) {
	printf("#fork failed, sorry\n");
	s->mode = 'C';
	finish_edit(s);
	if(!waitforeditor) return;
    } else {
	s->pid = waiting_for_pid;
	if(!waitforeditor) return;
	flush_output_buffer();
	/* childsig() will call finish_edit */
	while(waiting_for_pid) pause();
	lastline();  fputs(clreoln, stdout);
    }
    signal(SIGINT, intrsig);
    set_terminal();
}

static void finish_edit(s)
editsess *s;
{
    union {
	char buf[BUFSIZE];
	struct stat st;
    } b;
    editsess **sp;
    unsigned long size;
    int i, back_data_sent = data_sent;
#ifndef OMIT_SESSIONS
    int back_fd = sockfd;
    sockfd = main_sess(); 
#endif
    if(s->mode == 'E' && s->fname)
	if((s->fd = open(s->fname, O_RDONLY)) < 0 || fstat(s->fd, &b.st) < 0 ||
		((size = b.st.st_size) == s->size && b.st.st_mtime >= s->time))
	    s->mode = 'C';
	else {
	    /* file was changed by editor: send back result to server */
	    sprintf(b.buf, "_%lu_", s->key);
	    sprintf(b.buf, "%sE%lu\nE%lu\n",
		    MPI, strlen(b.buf) + size, s->key);
	    send_to_host(b.buf, strlen(b.buf));
	    for( ; size; size -= i) {
		i = sizeof(b.buf) < size ? sizeof(b.buf) : size;
		if((i = read(s->fd, b.buf, i)) > 0)
		    send_to_host(b.buf, i);
		else
		    syserr("read temp");
	    }
	    sprintf(footnote, "#session %d completed", s->num);
	}
    if(s->mode == 'C' && s->key) {		/* cancel editing session */
	sprintf(b.buf, "_%lu_", s->key);
	sprintf(b.buf, "%sE%u\nC%lu\n", MPI, (int)strlen(b.buf), s->key);
	send_to_host(b.buf, strlen(b.buf));
	sprintf(footnote, "#session %d cancelled", s->num);
    }
    data_sent = back_data_sent;		/* MP messages aren't prompted	*/
    if(s->fname) {
	close(s->fd);
	unlink(s->fname);
	free(s->fname);
    }
    for(sp = &edit_sess; *sp; sp = &(*sp)->next)
	if(*sp == s) { *sp = s->next;  break; }
    free((char*)s);
    /* parse (send) the edit_end message if this is the last editor... */ 
    if(!edit_sess && edit_end)
	parse(edit_end, 0, NULL);
#ifndef OMIT_SESSIONS
    sockfd = back_fd;
#endif
}

/*
 * cancel an editing session; does not free anything
 * (the child death signal handler will do that)
 */
void cancel_edit(s)
editsess *s;
{
    int sig = s->mode == 'C' ? SIGKILL : SIGTERM;
    s->mode = 'C';
    if(s->pid && kill(s->pid, sig) < 0)		/* Editicide */
	syserr("kill editor child");
}

/*
 * Our child has snuffed it.
 * If it was an editor/viewer, finish the session.
 */
void childsig()
{
    static int result;
    editsess *sp;
    int pid;
    while((pid = waitpid(-1, &result, WNOHANG)) > 0) {
	if(pid == waiting_for_pid) {
	    waiting_for_pid = 0;
	    wait_result = result;
	}
	for(sp = edit_sess; sp; sp = sp->next)
	    if(sp->pid == pid) {
		if(result) sp->mode = 'C';
		finish_edit(sp);
		break;
	    }
    }
    signal(SIGCHLD, childsig);
}
