/*
 * cancan  --  mud client with telnet protocol
 *
 * written by Mattias Engdegrd (Yorick) (f91-men@nada.kth.se) 1992-95
 * Maintainers: Ilie (finnag@pvv.unit.no) (and Yorick)
 * originally inspired by the Tintin client by Peter Unold but contains no
 * Tintin code.
 * Also greatly improved upon by Vivriel and Thuzzle. As usual, all the
 * developers are in debt to countless users for suggestions and debugging.
 *
 * This program is in the public domain.
 */

#define VERSION "2.6.2 (Yorick, Vivriel, Thuzzle, Ilie 1992-96)"
#define SAVEFILEVER 2

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#ifdef AIX
#include <sys/select.h>
#endif
#include <fcntl.h>
#include <signal.h>
#include <time.h>

#include "cancan.h"
#include "cmd.h"
#include "term.h"
#include "list.h"
#include "beam.h"

int select();
int pause();
int strncasecmp();
int setenv();
void bzero();
void unsetenv();

extern int errno, sys_nerr;
extern char *sys_errlist[];


/* 	function declarations		*/
static void init_env();
static void init_def_file();
static void mainlupe();
#ifdef MOTDFILE
static void printmotd();
#endif
static void get_remote_output();
#ifndef OMIT_SESSIONS
static void process_sub_input();
#endif
static void process_input();
static void getuserinput();
static void setsigs();
static void def_error();
static void read_settings();
static void flashparen();
void termsig();
void pipesig();
void denewbiefy();

/*	GLOBALS		*/

char *hostname;
int portnumber;
char *initstr;		/* initial command to run (send) on connect */

FILE *capfile;		/* capture file or NULL */

aliasnode *aliases;	/* head of alias list */
actionnode *actions;	/* head of action list */
marknode *markers;	/* head of mark list */
keynode *keydefs;	/* head of key binding list */

#ifndef OMIT_SESSIONS
struct consess {
  int active;
  char *id;
} con_sess[MAXFDSCAN];	/* connections indexed by socket descriptors */
#endif

int sockfd = 1;		/* current socket file descriptor */
fd_set fdset;		/* set of descriptors to select() on */

int waiting_for_pid;	/* = pid, when executing #! or editor */
int wait_result;	/* exit status of the process we waited for */

int linemode;		/* line mode flags (LM_* in cancan.h) */

/* booleans */
int mp_active;		/* 1 = we have #identified ourselves */
int new_settings;	/* set this when user defines new alias/action/etc */
int verbose;		/* 1 if command being parsed is user-entered */
int wrap = 1;		/* 1 = word wrap active */
int speedwalk;		/* speedwalk on? */
int verbatim;		/* 1 = don't expand aliases or process semicolons */
int flashback;		/* cursor is on excursion, put back there */
int edattr;		/* 1 = input line attributes are in effect */

/* for line editing */
void (*last_edit_func)();
int edlen;		/* length of current input line */
int pos;		/* cursor position in line */
char *prefixstr;	/* inserted in the editing buffer each time */
char edbuf[LINELEN]; 	/* line editing buffer */
char inserted_next[LINELEN];	/* inserted in buffer just once */

char footnote[LINELEN];	/* displayed at the bottom (just once) */
char promptstr[LINELEN];/* entire prompt string (for redrawing) */
char edattrbeg[CAPLEN];	/* starting input line attributes */
char edattrend[CAPLEN]; /* ending input line attributes */

int main(argc, argv)
int argc; char *argv[];
{
    if(argc < 2 || argc > 3) {
	fprintf(stderr, "usage:\t%s host port\nor\t%s definition-file\n",
		argv[0], argv[0]);
	return 1;
    }
    settermcap();
    default_bindings();
    init_env();
#ifndef OMIT_PRINTVER
    printver();
#endif
#ifdef MOTDFILE
    printmotd();
#endif
    if(argc == 2) {	/* one arg, assuming it is definition file name */
	init_def_file(argv[1]);
    } else {	/* two args: assume they are hostname and port number */
	hostname = argv[1];
	portnumber = atoi(argv[2]);
	keypad_bindings();
    }
    default_completions();
    set_terminal();
    setsigs();
    mainlupe();
    return 1;
}

/*
 * initialize environment variables
 */
static void init_env()
{
    char *p;
#ifdef NOSETENV
    int i = 0;
    extern char **environ;
    while(environ[i]) environ[i++] = str_dup(environ[i]);
#endif
    if((p = getenv("CANCANEDITOR"))) {
	setenv("EDITOR", p, 1);
	unsetenv("CANCANEDITOR");
    }
    if((p = getenv("CANCANPAGER"))) {
	setenv("PAGER", p, 1);
	unsetenv("CANCANPAGER");
    }
}

/*
 * initialize cancan from definition file or create one
 */
static void init_def_file(df)
char *df;
{
    char buf[LINELEN];
    char *p, *deff = df;
    /*
     * rules: if cancan dir is set it is searched first.
     *        if file doesn't exist, it is created there.
     * If a slash appears in the name, the cancan dir isn't used.
     */
    if(!strchr(df, '/') && (p = getenv("CANCANDIR"))) {
	strcpy(buf, p);
	if(buf[0] && buf[strlen(buf) - 1] != '/')
	    strcat(buf, "/");
	strcat(buf, df);
	if(access(buf, R_OK | W_OK) == 0 || access(df, R_OK | W_OK) != 0)
	    deff = buf;
    }
    setenv("SAVEFILE", deff, 1);
    if(access(deff, R_OK | W_OK) != 0) {
	printf("Creating %s\nHost name  :", deff);
	fflush(stdout);  fgets(buf, sizeof(buf), stdin);
	if(*skipspace(buf) != '\n') {
	    strtok(buf, "\n");  hostname = str_dup(buf);
	    printf("Port number:");
	    fflush(stdout);  fgets(buf, sizeof(buf), stdin);
	    portnumber = atoi(buf);
	}
	keypad_bindings();
	save_settings();
    } else
	read_settings(deff);
}

/*
 * show current version
 */
void printver()
{
    printf(
#if __STDC__
	"cancan ver " VERSION "\nCompiled " __DATE__ " (%s %s)\n",
#else
	"cancan ver %s (%s %s)\n", VERSION,
#endif
#ifdef VT100
	"vt100-only",
#else
	"termcaps",
#endif
#ifdef USETERMIO
	"termio"
#else
	"BSD sgtty"
#endif
    );
}

#ifdef MOTDFILE
/*
 * print the message of the day if present
 */
static void printmotd()
{
    char line[LINELEN];
    FILE *f = fopen(MOTDFILE, "r");
    if(f) {
	while(fgets(line, LINELEN, f))
	    fputs(line, stdout);
    }
}
#endif

/*
 * main loop.
 */
static void mainlupe()
{
    fd_set readfds;
    int err;
    struct timeval timeoutbuf, *timeout;
    
    FD_ZERO(&fdset);
    FD_SET(0, &fdset);		/* standard input (terminal) */
#ifndef OMIT_SESSIONS
    if(hostname && connect_sess(NULL, initstr, NULL, 0) < 0)
	exitcancan("");
#else
    if(!hostname || (sockfd = doconnect(hostname, portnumber)) < 0)
	exitcancan("");
    FD_SET(sockfd, &fdset);	/* add the socket to select() set */
    if(initstr) parse(initstr, 0, NULL);
#endif
    
    printf("Type #help for help.\n");
    
    /* make sure the input line is always the bottom one */
    winchsig();
    lastline();
    
    /*		main lupe	*/
    
    for(;;) {
	flush_output_buffer();	/* sploooosh */
	fflush(stdout);		/* ditto */

	readfds = fdset;
	timeout = 0;
	if(flashback) {
	    timeoutbuf.tv_sec = 0;
	    timeoutbuf.tv_usec = FLASHDELAY;
	    timeout = &timeoutbuf;
	}

	err = select(MAXFDSCAN, &readfds, NULL, NULL, timeout);

	if(flashback) { moveto(flashback);  flashback = 0; }
	if(*footnote) clear_footnote();
	if(err < 0) {
	    if(errno != EINTR) syserr("select");
	} else {
#ifndef OMIT_SESSIONS
	    if (FD_ISSET(0, &readfds)) {
		err--;
		sockfd = main_sess();
		getuserinput();
	    }
	    for (sockfd = 3; err && (sockfd < MAXFDSCAN); sockfd++)
		if (FD_ISSET(sockfd, &readfds)) {
		    err--;
		    get_remote_output();
		}
#else
	    if (FD_ISSET(0, &readfds))
		getuserinput();
	    if (FD_ISSET(sockfd, &readfds))
		get_remote_output();
#endif
	}
	if(*footnote) show_footnote();
    }
}

/*
 * get output from the socket and process/display it.
 */
static void get_remote_output()
{
    static char buffer[BUFSIZE];
    int i, got = read_socket(buffer, sizeof(buffer) - 1);
    if(!got) return;		/* leave room  ^^^ for terminating zero */
#ifdef DEBUGCODE
    /* debug code to see in detail what strange codes come from the server */
    printf("{");
    for(i = 0; i < got; i++) 
	printf(isprint(buffer[i]) ? "%c" : "[%d]", buffer[i]);
    printf("}\n");
#endif
  
#ifndef OMIT_SESSIONS
    if(!con_sess[sockfd].active) return;  /* process only active sessions */
#endif
    if(linemode & LM_CHAR) {
	while((i = write(1, buffer, got)) >= 0 ? (got -= i) : errno == EINTR);
	if(i < 0) syserr("write");
    } else {
	/* line-at-a-time mode: process input in a number of ways */
	clearinpline();
	if(edattr) { edattr = 0; fputs(edattrend, stdout); }
#ifndef OMIT_SESSIONS
	if(!is_main_sess()) process_sub_input(buffer, got);
	else
#endif
	process_input(buffer, got);
	redraw_input();
    }
}

#ifndef OMIT_SESSIONS
/*
 * process input from remote host (subsidiary session):
 * assert(sizeof(buf) >= size + 1)
 */
static void process_sub_input(buf, size)
char *buf; int size;
{
    char *line, *p;
    int i;

    if(col0) printf("%s%s", begoln, clreoln);
    for(i = 0; i < size; i++) {
	if(buf[i] == '\r') continue;
	for(p = line = buf + i; i < size && buf[i] != '\n'; i++)
	    if(buf[i]) *p++ = buf[i];	/* cut out NUL's */
	*p = '\0';		/* there is room */
	/* FEATURE: no wrapping, no marking in subsidiary session */
	/* BUG: escape codes not removed */
	if(search_action(line)) {
	    printf("##%s> %.*s", con_sess[sockfd].id,
		    cols_1 - 4 - (int)strlen(con_sess[sockfd].id), line);
	    newline();
	}
    }
    check_prompt();
}
#endif

/*
 * process input from remote host:
 * detect special sequences, trigger actions, locate prompt, word-wrap,
 * print to stdout
 * assert(sizeof(buf) >= size + 1)
 */
static void process_input(buf, size)
char *buf; int size;
{
    static editsess *esess;
    char *p;
    int i = 0;

    if(data_sent) {
	data_sent = 0;
	if(esess) esess->mode = 'C';	/* sending cancels the session */
    }
    while(i < size) {
	p = buf + i;
	if(esess || (mp_active && !col0 && !strncmp(p, MPI, strlen(MPI))))
	    i += mp_receive(&esess, p, size - i);
	else {
	    /* process block one line at a time: */
	    if(!col0 && *p == '\r') i++;		/* leading CR */
	    for(p = buf; i < size && buf[i] != '\n'; i++)
		if((*p = buf[i])) p++;			/* cut out NUL's */
	    if(p != buf && *(p - 1) == '\r') p--;	/* trailing CR */
	    *p = '\0';
	    if(capfile) fputs(buf, capfile);
	    if(i++ < size) {
		if(capfile) putc('\n', capfile); 
		if(!search_action(buf)) continue;	/* line #gagged */
		promptstr[0] = '\0';
		if(buf[0])
		    if(wrap) wrap_out(buf);
		    else mark_and_print(buf);
		promptstr[0] = '\0';
		col0 = 0;  newline();
	    } else {
		search_action(buf);
		if(!(p = wrap_out(buf))) p = buf;
		else promptstr[0] = '\0';
		if(col0 < cols_1 && strlen(promptstr) + strlen(p) < LINELEN)
		    strcat(promptstr, p);
		else {
		    col0 = 0;  newline();
		}
		break;
	    }
	}
    } /* end while */
}

/*
 * read terminal input and send to parser.
 * decode keys that send escape sequences
 */
static void getuserinput()
{
    int i;
    unsigned char c;
    static char typed[CAPLEN];	/* chars typed so far (with partial match) */
    static int nchars = 0;	/* number of them */
    keynode *p;

    /* We have 4 possible line modes:
     * line mode, local echo: line editing functions in effect
     * line more, no echo: sometimes used for passwords, no line editing
     * char mode, no echo: send a character directly, no local processing
     * char mode, local echo: extremely rare, do as above.
     */
    if(read(0, &c, 1) != 1)
	syserr("read from stdin");
    if(linemode & (LM_CHAR | LM_NOECHO)) {
	if(linemode & LM_CHAR) {
	    while((i = write(sockfd, &c, 1)) < 0 && errno == EINTR);
	    if(i != 1) syserr("write to socket");
	    if(!(linemode & LM_NOECHO))
		putchar(c);
	} else {
	    /* sending password (line mode, no echo) */
	    edbuf[edlen++] = c;
	    if(edbuf[0] == '\003') intrsig();
	    if((c == '\n' && (newline(), 1)) || edlen >= LINELEN - 2) {
		send_to_host(edbuf, edlen);
		edlen = 0;
	    }
	    edbuf[pos = edlen] = 0;
	}
    } else {
	/* normal mode (echo, line buffered) */
#ifndef VT100
	if(metakey && (c & 0x80)) {	/* did meta key set the high bit? */
	    typed[nchars++] = '\033';	/* prepend esc */
	    c &= 0x7f;
	}
#endif
	if(!nchars && isprint(c)) {
	    if(edlen >= LINELEN - 1) return;
	    if(!edattr && *edattrbeg) {
		edattr = 1;  fputs(edattrbeg, stdout);
	    }
	    insertchars(&c, 1);
	    if(isrparen(c)) flashparen(c);
	    last_edit_func = 0;
	} else {
	    /* search function key strings for match */
	    typed[nchars++] = c;
	    if((p = keydefs))
		while(strncmp(typed, p->sequence, nchars) && (p = p->next));
	    if(p) {
		if(p->sequence[nchars]) return;		/* partial match */
		p->funct(p->call_data);
		last_edit_func = p->funct;
	    }
	    nchars = 0;
	}
    }
}

/*
 * Flash cursor at parentheses that matches c inserted before current pos
 */
static void flashparen(c)
unsigned char c;
{
    int lev = 0, i = pos;
    while(--i >= 0)
	if(isrparen(edbuf[i])) {
	    lev++;
	} else if(islparen(edbuf[i]) && !--lev) {
	    if(LPAREN(c) == edbuf[i]) {
		flashback = pos;
		moveto(i);
	    }
	    break;
	}
}

#ifdef NEWBIE
/*
 * Some people just can't write prose. It's horrible, but here is the
 * fix we have been waiting for.
 */
void denewbiefy()
{
  char *p, *q, *r;
  char buf[LINELEN];
  int n;
  
  strcpy(buf, edbuf);
  p = buf; r = edbuf;
  while(*p) 
   {
     while(*p && !islabel(*p)) *r++ = *p++;
     q = p;
     while(*p && islabel(*p)) p++;
     if(*p || *q) 
      {
	n = p - q;
	if(strncasecmp(q, "u", n) == 0) 
	 {
	   strcpy(r, "you"); r += 3;
	 }
	else if(strncasecmp(q, "ur", n) == 0) 
	 {
	   strcpy(r, "your"); r += 4;
	 }
	else if(strncasecmp(q, "urs", n) == 0) 
	 {
	   strcpy(r, "yours"); r += 5;
	 }
	else if(strncasecmp(q, "r", n) == 0) 
	 {
	   strcpy(r, "are"); r += 3;
	 }
	else if(strncasecmp(q, "4", n) == 0) 
	 {
	   strcpy(r, "for"); r += 3;
	 }
	else if(strncasecmp(q, "2", n) == 0) 
	 {
	   strcpy(r, "to"); r += 2;
	 }
	else if(strncasecmp(q, "b", n) == 0) 
	 {
	   strcpy(r, "be"); r += 2;
	 } 
	else if(strncasecmp(q, "b4", n) == 0) 
	 {
	   strcpy(r, "before"); r += 6;
	 }
	else 
	 {
	   strcpy(r, q); r += n;
	 }
      }
   }
  *r = '\0';
  edlen = r - edbuf;
}
#endif

/*
 * terminate cancan as cleanly as possible
 */
void exitcancan(message)
char *message;
{
    static int again;
    if(!again) {
	again = 1;
	reset_terminal();
	printf("%s%s#cancan stopped dancing.\n", edattrend, message);
	exit(0);
    }
}
 
#ifndef NOSHELL
/*
 * execute a shell command and return the result
 */
int execute(arg)
char *arg;
{
    flush_output_buffer();
    fflush(stdout);
    signal(SIGINT, SIG_IGN);
    if(!(waiting_for_pid = fork())) {
	static char *args[4] = { "/bin/sh", "-c", NULL, NULL };
	args[2] = arg;
	execvp(args[0], args);
	_exit(1);
    } else if(waiting_for_pid < 0) {
	printf("#fork failed, sorry\n");
    } else {
	while(waiting_for_pid) pause();
    }
    signal(SIGINT, intrsig);
    return wait_result;
}
#endif

/*
 * set up our signal handlers
 */
static void setsigs()
{
  signal(SIGCHLD, childsig);
  signal(SIGINT, intrsig);
  signal(SIGTERM, termsig);
  signal(SIGPIPE, pipesig);
  signal(SIGWINCH, winchsig);
}

void intrsig()
{
  if(sockfd > 1) close(sockfd);
  exitcancan("\n#interrupted.\n");
}

void pipesig()
{
  exitcancan("\n#broken pipe.\n");
}

void termsig()
{
  exitcancan("\n#termination signal.\n");
}

/*
 * suspend ourselves
 * return -1 if no job control
 */
void suspend_cancan()
{
    if(signal(SIGTSTP, SIG_IGN) == SIG_DFL) {
	if(edattr) { edattr = 0;  fputs(edattrend, stdout); }
	reset_terminal();  fflush(stdout);
	signal(SIGINT, SIG_IGN);
	signal(SIGTERM, SIG_IGN);
	signal(SIGTSTP, SIG_DFL);
	kill(0, SIGTSTP);
	setsigs();
	set_terminal();
	winchsig();		/* in case size changed meanwhile */
	lastline();		/* so we know where the cursor is */
	check_prompt();
	redraw_input();
    } else
	strcpy(footnote, "#your shell probably lacks job control.");
}

/*
 * print an error message
 */
int syswarn(msg)
char *msg;
{
    if(errno > 0 && errno < sys_nerr)
	printf("%s (%s)\n", msg, sys_errlist[errno]);
    else
	printf("%s\n", msg);
    return -1;
}

/*
 * print system call error message and terminate
 */
void syserr(msg)
char *msg;
{
    reset_terminal();
    fprintf(stderr, "\n#fatal system call error: %s (", msg);
    if(errno > 0 && errno < sys_nerr)
	fprintf(stderr, "%s)\n", sys_errlist[errno]);
    else
	fprintf(stderr, "%d)\n", errno);
    exit(1);
}


/*
 * redisplay the prompt if needed
 */
void check_prompt()
{
    if(data_sent) {
	promptstr[0] = col0 = 0;
    } else if(col0) {
	if(edattr) { edattr = 0; fputs(edattrend, stdout); }
	mark_and_print(promptstr);
    }
}

static void def_error(ln)
int ln;
{
    static int n_err = 0;
    if(n_err++ >= 10) exit(1);
    printf("bad definition file format, line %d\n", ln);
}

/*
 * read definitions from file
 */
static void read_settings(deff)
char *deff;
{
    char buf[2 * LINELEN];
    FILE *f = fopen(deff, "r");
    char *p, *cmd;
    int ln;

    if(!f) syserr("open definition file");
    for(ln = 1; fgets(buf, sizeof(buf), f); ln++) {
	if(*buf == '\n') continue;
	for(p = cmd = buf; *p && *p != ' '; p++);
	if(*p && *cmd++ == '#') {
	    *p++ = 0;
	    strtok(p, "\n");	/* strip lf */
	    if(strcmp(cmd, "savefile") == 0) {
		if(atoi(p) > SAVEFILEVER)
		    printf("Warning: this cancan version is too old!\n");
	    } else if(strcmp(cmd, "alias") == 0) {
		parse_alias(p);
	    } else if(strcmp(cmd, "host") == 0) {
		if(!(p = strtok(p, " ")))
		    def_error(ln);
		else {
		    str_assign(&hostname, p);
		    if((p = strtok(NULL, " ")))
			portnumber = atoi(p);
		}
	    } else if(strcmp(cmd, "init") == 0) {
		str_assign(&initstr, p);
	    } else if(strcmp(cmd, "action") == 0) {
		parse_action(p);
	    } else if(strcmp(cmd, "mark") == 0) {
		parse_marker(p);
	    } else if(strcmp(cmd, "bind") == 0) {
		parse_bind(p);
	    } else {
		def_error(ln);
	    }
	} else
	    def_error(ln);
    }
    fclose(f);
}

/*
 * save settings in definition file
 */
void save_settings()
{
    char buf[LINELEN], *cp;
    FILE *f;
    aliasnode *al;
    actionnode *ac;
    marknode *m;
    keynode *k;
  
    if(!(cp = getenv("SAVEFILE"))) return;
    if(!(f = fopen(cp, "w"))) {
	printf("#error: can't write to definition file %s\n", cp);
	return;
    }
    fprintf(f, "#savefile %d\n", SAVEFILEVER);
    if(hostname)
	fprintf(f, "#host %s %d\n", hostname, portnumber);
    if(initstr)
	fprintf(f, "#init %s\n", initstr);
    for(al = aliases; al; al = al->next)
	fprintf(f, "#alias %s=%s\n", al->name, al->subst);
    for(ac = actions; ac; ac = ac->next) {
	escape_specials(ac->pattern, buf, "\\\"");
	fprintf(f, "#action %c%s \"%s\"=%s\n", (ac->active) ? '+' : '-',
		ac->label, buf, ac->commands);
    }
    for(m = markers; m; m = m->next) {
	escape_specials(m->pattern, buf, "\\=");
	fprintf(f, "#mark %s=%s\n", buf, attr_name(m->attrcode));
    }
    /* FEATURE: deleted default keys should be saved too, as "#bind <key>=" */
    for(k = keydefs; k; k = k->next)
	if(k->call_data) {
	    escape_seq(k->sequence, buf);
	    fprintf(f, "#bind %s %s=%s\n", k->name, buf, k->call_data);
	}
    fclose(f);
}

/*
 * if alias recording is active, record given line in that alias
 */
void record_line(line)
char *line;			/* mustn't be trashed! */
{
    aliasnode *np;
    if(!strcmp(line, recording_name)) return;
    if(!(np = *lookup_alias(recording_name))) {
	add_aliasnode(recording_name, line);
    } else {
	char new_line[LINELEN];
	int room = LINELEN - 1 - strlen(line);
	if(strlen(np->subst) >= room) {
	    /* too long, must be split */
	    int i = 0;
	    do {
	    	/* try name1, name2, ... */
		sprintf(new_line, "%s%d", recording_name, ++i);
	    } while(*lookup_alias(new_line));
	    add_aliasnode(new_line, np->subst);
	} else
	    strcpy(new_line, np->subst);
	if(strlen(new_line) < room) {
	    strcat(new_line, ";");
	    strcat(new_line, line);
	    str_assign(&np->subst, new_line);
	}
    }
}

char *str_dup(s)
char *s;
{
    char *k = (char *)malloc(strlen(s) + 1);
    if(!k) syserr("malloc");
    return strcpy(k, s);
}

void str_assign(str, s)
char **str, *s;
{
    if(*str) free(*str);
    *str = s ? str_dup(s) : 0;
}

#ifndef OMIT_SESSIONS
/*
 * Below are multiple-session support functions:
 *
 * note on con_sess[MAXFDSCAN]:
 *	con_sess[0].active holds total no. of connections
 *	con_sess[1].active holds main socket descriptor
 */

/*
 * return TRUE if current session being processed is main connection
 */
int is_main_sess ()
{
  return (sockfd == con_sess[1].active);
}

/*
 * return main session sockfd
 */
int main_sess ()
{
  return con_sess[1].active;
}

/*
 * return session's index or socket descriptor given id,
 * or 0 if null or invalid id is given
 */
int find_sess (id, notify)
char *id;
int notify;
{
  int i, n;
  
  for (i=3, n=con_sess[0].active; n; i++)
     if (con_sess[i].id) 
      {
	n--;
	if (!strcmp(con_sess[i].id, id))
	   return i;
      }
  if (notify)
     printf("#no such connection\n");
  return 0;
}

/*
 * connect another session
 */
int connect_sess (id, init, host, port)
char *id;	/* session id - NULL if main session */
char *init;	/* optional init string */
char *host;
int port; 
{
    int fd;
    if((!host && !(host = hostname)) || (!port && !(port = portnumber))) {
	printf("#please specify the hostname and port number.\n");
	return -1;
    }
    if(init && !*init) init = 0;
    /* dial the number by moving the right index in small circles */
    if((fd = doconnect(host, port)) < 0)
	return -1;
    FD_SET(fd, &fdset);		/* add socket to select() set */
    con_sess[0].active++;
    if(!con_sess[1].active)	/* save main socket fd */
	con_sess[1].active = sockfd = fd;
    if(!id) {			/* main session */
	con_sess[fd].active = 1;
	con_sess[fd].id = "main";
    } else {
	con_sess[fd].active = !init;
	con_sess[fd].id = str_dup(id);
    }
    if(init) parse(init, 0, NULL);
    return fd;
}

/*
 * disconnect session
 */
void disconnect_sess (id)
char *id;
{
  int sfd;
  
  if (id)
   {  /* #disconnect cmd */
     if (!(sfd = find_sess(id, 1)))
     	return;
     close(sfd);
   } 
  else
     sfd = sockfd;  /* connection closed by remote host */
  
  printf("%s%s#connection \"%s\" closed.\r\n",
	begoln, clreoln, con_sess[sfd].id);
  if (sfd == con_sess[1].active)  /* main connection closed */
    exitcancan("");
  
  check_prompt();	/* prompt is redisplayed there */
  FD_CLR(sfd, &fdset);
  con_sess[0].active--;
  con_sess[sfd].active = 0;
  free(con_sess[sfd].id);
  con_sess[sfd].id = 0;
}

/*
 * toggle output display from another session
 */
void snoop_sess (id)
char *id;
{
    int sfd;
  
    sfd = find_sess(id, 1);
    if(sfd)
	con_sess[sfd].active ^= 1;
}

/*
 * show open connections
 */
void show_connections()
{
    int i, n;
    if(!(n = con_sess[0].active)) {
	printf("#No connections have been made\n");
    } else {
	printf("#The following connection%s open:\n",
		(n == 1) ? " is" : "s are");
	for(i = 3; n; i++)
	    if(con_sess[i].id) {
		n--;
		printf("##%-8s %s\n", con_sess[i].id, !con_sess[i].active ?
			"" : i == con_sess[1].active ?
			"(active)" : "(snooped)");
	    }
    }
}
#endif

#ifdef NOSETENV
/*ARGSUSED*/
int setenv(variable, value, overwrite)
char *variable, *value;
int overwrite;
{
    extern char **environ;
    char buf[LINELEN], *k;
    int i, len = strlen(variable);
    sprintf(buf, "%s=%s", variable, value);
    for(i = 0; (k = environ[i]); i++)
	if(!strncmp(k, variable, len) && (k[len] == '=' || !k[len])) {
	    str_assign(environ + i, buf);
	    return 1;
	}
    putenv(str_dup(buf));
    return 1;
}

void unsetenv(variable)
char *variable;
{
    extern char **environ;
    int i, len = strlen(variable);
    char *k;
    for(i = 0; (k = environ[i]); i++)
	if(!strncmp(k, variable, len) && (k[len] == '=' || !k[len])) {
	    free(k);
	    while((environ[i] = environ[i + 1])) i++;
	}
}
#endif
