/*
 * edit.c
 * line editing functions for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include "cancan.h"
#include "cmd.h"
#include "term.h"

int strncasecmp();

static void prev_line(), next_line(), prev_char(), next_char();
static void next_word(), prev_word(), del_word_right(), del_word_left();
static void del_char_right(), del_char_left();
static void end_of_line(), beginning_of_line();
static void kill_to_eol(), redraw_input_line();
static void transpose_chars(), transpose_words();
static void enter_into_history(), enter_line();
static void complete_word(), complete_line();
static void matching_bracket();

static void put_in_history();
static void swap_history();
static void demote_word();
static void overtype();
static int skip_word();
static int skip_nonword();

/* completion list */
static int wordindex;
static char *words[WORDSKEPT];
static char wordflags[WORDSKEPT];
#if (WORDSKEPT - 1) >> 8
static unsigned int wordnext[WORDSKEPT];
static unsigned int wordprev[WORDSKEPT];
#else
static unsigned char wordnext[WORDSKEPT];
static unsigned char wordprev[WORDSKEPT];
#endif

/* history buffer */
static char *hist[MAXHIST];	/* saved history lines */
static int curline;		/* current history line */
static int pickline;		/* line to pick history from */

struct defbind_node defbinds[] = {
  { "enter",	"\n",		NULL,			enter_line },
  { "return",	"\r",		NULL,			enter_line },
  { "bs",	"\b",		"&del-char-left",	del_char_left },
  { "del",	"\177",		"&del-char-left",	del_char_left },
  { "tab",	"\t",		"&complete-word",	complete_word },
  { "left",	key_left,	"&prev-char",		prev_char },
  { "right",	key_right,	"&next-char",		next_char },
  { "up",	key_up,		"&prev-line",		prev_line },
  { "down",	key_down,	"&next-line",		next_line },
  { "C-a",	"\001",		"&beginning-of-line",	beginning_of_line },
  { "C-b",	"\002",		"&prev-char",		prev_char },
  { "C-c",	"\003",		NULL,			intrsig },
  { "C-d",	"\004",		"&del-char-left",	del_char_right },
  { "C-e",	"\005",		"&end-of-line",		end_of_line },
  { "C-f",	"\006",		"&next-char",		next_char },
  { "C-g",	"\007",		"&matching-bracket",	matching_bracket },
  { "C-k",	"\013",		"&kill-to-eol",		kill_to_eol },
  { "C-l",	"\014",		"&redraw-input-line",	redraw_input_line },
  { "C-n",	"\016",		"&next-line",		next_line },
  { "C-p",	"\020",		"&prev-line",		prev_line },
  { "C-t",	"\024",		"&transpose-chars",	transpose_chars },
  { "C-y",	"\031",		"&enter-into-history",	enter_into_history },
  { "C-z",	"\032",		NULL,			suspend_cancan },
  { "M-tab",	"\033\t",	"&complete-line",	complete_line },
  { "M-b",	"\033b",	"&prev-word",		prev_word },
  { "M-d",	"\033d",	"&del-word-right",	del_word_right },
  { "M-f",	"\033f",	"&next-word",		next_word },
  { "M-t",	"\033t",	"&transpose-words",	transpose_words },
  { "M-del",	"\033\177",	"&del-word-left",	del_word_left },
  { "",		NULL,		NULL,			NULL }
};

/*
 * redraw the input line
 */
static void redraw_input_line()
{
    clearinpline();
    redraw_input();
}

/*
 * transpose two characters to the left
 */
static void transpose_chars()
{
    if(edlen > 1) {
	char c;
	int k = pos;
	if(k && k-- >= edlen) k--;
	moveto(k);
	c = edbuf[k];
	overtype(edbuf[k + 1]);
	overtype(c);
    }
}

/*
 * transpose current word and the previous one
 */
static void transpose_words()
{
    int a, b, c;
    prev_word();
    a = skip_word(skip_nonword(pos));
    b = skip_nonword(a);
    c = skip_word(b);
    if(b < c) {
	char buf[LINELEN];
	sprintf(buf, "%.*s%.*s%.*s",
		c - b, edbuf + b, b - a, edbuf + a, a - pos, edbuf + pos);
	for(a = 0; buf[a]; overtype(buf[a++]));
    }
}

/*
 * erase everything to the end of line
 */
static void kill_to_eol()
{
    int cline = CURLINE(pos);
    fputs(clreoln, stdout);
    if(CURLINE(edlen) > cline) {
	printf("\n%s", clreoscr);
	optmove(cline + 1, 0, cline, CURCOL(pos));
    }
    edbuf[edlen = pos] = '\0';
}

/*
 * move cursor to end of line
 */
static void end_of_line()
{
    moveto(edlen);
}

/*
 * move cursor to beginning of line
 */
static void beginning_of_line()
{
    moveto(0);
}

/*
 * delete a character to the right
 */
static void del_char_right()
{
    if(pos < edlen)
	deletechars(1);
}

/*
 * delete a character to the left
 */
static void del_char_left()
{
    if(pos) {
	moveto(pos - 1);
	deletechars(1);
    }
}

/*
 * move the editline into history
 */
static void enter_into_history()
{
    clearinpline();
    put_in_history(edbuf);
    pickline = curline;
    edbuf[0] = '\0';
    pos = edlen = 0;
}

/*
 * enter a line
 */
static void enter_line()
{
    char *p;
    int ls = (pos >= edlen) ? 1 : 1 + CURLINE(edlen) - CURLINE(pos);
    do newline(); while(--ls); 
    if(capfile) fprintf(capfile, "%s\n", edbuf);
#ifdef NEWBIE
    denewbiefy();
#endif
    inserted_next[0] = '\0';
    if(verbatim && edbuf[0] != '#') {
	edbuf[edlen] = '\n';
	send_to_host(edbuf, edlen + 1);
	edbuf[edlen] = '\0';
    } else {
	if(edattr) { edattr = 0; fputs(edattrend, stdout); }
	new_settings = 0;
        verbose = 1;
	parse(edbuf, 1, NULL);
	verbose = 0;
	if(new_settings)	/* user defined something -- save */
	    save_settings();
    }
    check_prompt();
    p = hist[curline ? curline - 1 : MAXHIST - 1];
    /* don't put identical lines in history, nor short ones */
    if(edlen > 1 && (!p || strcmp(edbuf, p)))
	put_in_history(edbuf);
    pickline = curline;
    strcpy(edbuf,
	    inserted_next[0] ? inserted_next :
	    prefixstr ? prefixstr : "");
    pos = edlen = strlen(edbuf);
    redraw_input();
}

/*
 * move one word forward
 */
static void next_word()
{
    moveto(skip_word(skip_nonword(pos)));
}

/*
 * move one word backward
 */
static void prev_word()
{
    int i = pos;
    while(i && !islabel(edbuf[i - 1])) i--;
    while(i && islabel(edbuf[i - 1])) i--;
    moveto(i);
}

/*
 * delete word to the right
 */
static void del_word_right()
{
    deletechars(skip_word(skip_nonword(pos)) - pos);
}

/*
 * delete word to the left
 */
static void del_word_left()
{
    int i = pos;
    prev_word();
    deletechars(i - pos);
}

/*
 * get previous line from history list
 */
static void prev_line()
{
    int i = pickline - 1;
    if(i < 0) i = MAXHIST - 1;
    swap_history(i);
}

/*
 * get next line from history list
 */
static void next_line()
{
    int i = pickline + 1;
    if(i >= MAXHIST) i = 0;
    swap_history(i);
}

/*
 * move one char backward
 */
static void prev_char()
{
    if(pos)
	moveto(pos - 1);
}

/*
 * move one char forward
 */
static void next_char()
{
    if(pos < edlen)
	moveto(pos + 1);
}
 
/*
 * move to matching bracket
 */
static void matching_bracket()
{
    static char brackets[] = "(){}[]<>";
    char c = edbuf[pos], *p = strchr(brackets, c);
    if(p) {
	int i = pos, nested = 0, backwd = ((p - brackets) & 1) ? -2 : 0;
	char k = *(p + backwd + 1);
	while(backwd ? i-- : ++i < edlen)
	    if(edbuf[i] == c) nested++;
	    else if(edbuf[i] == k && !nested--) {
		moveto(i);
		break;
	    }
    }
}

/*
 * execute string as if typed
 */
void key_run_command(cmd)
char *cmd;
{
    clearinpline();
    if(!edattr) fputs(edattrbeg, stdout);
    printf("%s%s\n", cmd, edattrend);  edattr = 0;
    if(capfile) fprintf(capfile, "%s\n", cmd);
    parse(cmd, 1, NULL);
    check_prompt();
    redraw_input();
}

/*
 * match and complete a word referring to the word list
 */
static void complete_word()
{
    static int index, comp_len, root_len;
    char *root, *p;
    int k, n;

    /* determine current index, root, root_len */
    if(last_edit_func == complete_word && (k = comp_len)) {
	moveto(pos - k);
	n = pos - root_len;
    } else {
	if((n = pos)) while(!strchr(DELIM, edbuf[n - 1]) && --n);
	k = 0;  index = wordindex;  root_len = pos - n;
    }
    root = edbuf + n;  comp_len = 0;

    /* look up next valid completion and print it */
    while((p = words[index = wordnext[index]]))
	if(!strncasecmp(p, root, root_len) && *(p += root_len) &&
		(n = strlen(p)) + edlen < LINELEN) {
	    comp_len = n;
	    while(k ? (k--, overtype(*p++), --n) : (insertchars(p, n), 0));
	    break;
	}
    if(k) deletechars(k);

    /* delete duplicate instances of the word */
    if(p && !(wordflags[k = index] & 1)) {
	wordflags[k]++;  p = words[k];  n = wordnext[k];
	while(words[k = n])
	    if(n = wordnext[k], !strcmp(p, words[k])) {
		demote_word(k);
		free(words[k]);  words[k] = 0;
		wordflags[index] |= wordflags[k];
		if((wordflags[k] &= 1)) break;
	    }
    }
}

/*
 * match and complete entire lines backwards in history
 */
static void complete_line()
{
    if(edlen) {
	int i = curline;
	while(--i != curline) {
	    if(i < 0) i = MAXHIST - 1;
	    if(hist[i] && !strncmp(hist[i], edbuf, edlen)) {
		clearinpline();
		strcpy(edbuf, hist[i]);
		pos = edlen = strlen(edbuf);
		redraw_input();
		break;
	    }
	}
    }
}

/*
 * put string in history at current position
 * (string is assumed to be trashable)
 */
static void put_in_history(str)
char *str;
{
    char *p;
    str_assign(hist + curline, str);
    if(++curline >= MAXHIST) curline = 0;
  
    /* split into words and put into completion list (unless too short) */
    for(p = strtok(str, DELIM); p; p = strtok(NULL, DELIM))
	if(strlen(p) >= 3) put_word(p);
}

/*
 * swap editline with a history line
 */
static void swap_history(line)
int line;
{
    if(hist[line]) {
	str_assign(hist + pickline, edbuf);
	pickline = line;
	clearinpline();
	strcpy(edbuf, hist[line]);
	pos = edlen = strlen(edbuf);
	redraw_input();
    }
}

/*
 * put word in word completion ring
 */
void put_word(s)
char *s;
{
    int r = wordindex;
    words[r] = str_dup(s);  wordflags[r] = 0;
    while(wordflags[r = wordprev[r]] & 2);	/* skip permanent entries */
    demote_word(r);  wordindex = r;
    if(words[r]) { free(words[r]);  words[r] = 0; }
}

/*
 * reinsert a node just before wordindex ie make it the last word
 */
static void demote_word(i)
int i;
{
    wordnext[wordprev[i]] = wordnext[i];
    wordprev[wordnext[i]] = wordprev[i];
    wordprev[i] = wordprev[wordnext[i] = wordindex];
    wordprev[wordindex] = wordnext[wordprev[wordindex]] = i;
}

/*
 * print a char on current position (overwrite), advance position
 */
static void overtype(c)
char c;
{
    edbuf[pos++] = c;
    putchar(c);
    if(!CURCOL(pos))
	putchar('\n');
}

/*
 * return the position after leading islabel() chars
 */
static int skip_word(i)
int i;
{
    while(islabel(edbuf[i])) i++;
    return i;
}

/*
 * return the position after leading word separator chars
 */
static int skip_nonword(i)
int i;
{
    while(i < edlen && !islabel(edbuf[i])) i++;
    return i;
}

/*
 * initialize the word completion ring
 */
void default_completions()
{
    char buf[LINELEN];
    cmdstruct *p;
    int i;
    for(i = 0, buf[0] = '#', p = commands; p->name; p++)
	if(p->funct && strlen(p->name) >= 3) {
	    if(++i >= WORDSKEPT) break;
	    strcpy(buf + 1, p->name);
	    words[i] = str_dup(buf);  wordflags[i] = 3;
	}
    for(i = WORDSKEPT; i--; wordprev[i] = i - 1, wordnext[i] = i + 1);
    wordprev[0] = WORDSKEPT - 1;  wordnext[WORDSKEPT - 1] = 0;
}
