/*
 * term.c - terminal handling routines for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>

#ifdef USETERMIO
#ifdef APOLLO
#include "/sys5.3/usr/include/sys/termio.h"
#else
/* including both termio.h and termios.h necessary on AIX? */
#include <termio.h>
#include <termios.h>
#endif
#else /* not USETERMIO */
#include <sys/ioctl.h>
#endif

#ifdef BSD_386
#include <sys/ioctl_compat.h>
#define O_RAW RAW
#define O_ECHO ECHO
#endif

#ifdef USETERMIO
#if defined(TCSETS) || defined(TCSETATTR)
/* should perhaps use TCSETSF? */
#ifndef TCSETS		/* cc for HP-UX  SHOULD define this... */
#define TCSETS TCSETATTR
#define TCGETS TCGETATTR
#endif
typedef struct termios termiostruct;
#else
/* should perhaps use TCSETAF? */
#define TCSETS TCSETA
#define TCGETS TCGETA
typedef struct termio termiostruct;
#endif
#endif /* USETERMIO */

#include "cancan.h"
#include "list.h"
#include "term.h"

int ioctl();

static char *str_mark();
static void strins();

#ifndef VT100

int tgetent();
int tgetnum();
int tgetflag();
char *tgetstr();
char *tgoto();

static int extract();

/* terminal escape sequences */
char kpadstart[CAPLEN], kpadend[CAPLEN];
char begoln[CAPLEN], clreoln[CAPLEN], clreoscr[CAPLEN];
char leftcur[CAPLEN], rightcur[CAPLEN], upcur[CAPLEN], curgoto[CAPLEN];
char delchar[CAPLEN], inschar[CAPLEN], insstop[CAPLEN];
/* attribute changers: */
char modebold[CAPLEN], modeinv[CAPLEN], modeuline[CAPLEN];
char modestandon[CAPLEN], modestandoff[CAPLEN], modenorm[CAPLEN];

char key_left[CAPLEN], key_right[CAPLEN];
char key_up[CAPLEN], key_down[CAPLEN];

static int len_begoln, len_clreoln;
static int len_leftcur, len_upcur, gotocost;
static int deletecost, insertcost, insertfinish;

int metakey;		/* terminal has meta key setting high bit */
int wrapglitch;		/* strange wrap behaviour (a la vt100) */

int cols, lines;	/* screen size */
int cols_1, lines_1;	/* for efficiency's sake, cols - 1, lines - 1 */

#else	/* hard-coded vt100 features if no termcap: */

char kpadstart[] = "", kpadend[] = "";
char begoln[] = "\r";
char clreoln[] = "\033[K", clreoscr[] = "\033[J";
char leftcur[] = "\033[D", rightcur[] = "\033[C", upcur[] = "\033[A";
char modebold[] = "\033[1m", modeinv[] = "\033[7m", modeuline[] = "\033[4m";
char modestandon[] = "", modestandoff[] = "", modenorm[] = "\033[m";

char key_left[] = "\033[D", key_right[] = "\033[C";
char key_up[] = "\033[A", key_down[] = "\033[B";

static int len_begoln = 1, len_leftcur = 3, len_upcur = 3, gotocost = 8;

#define wrapglitch (1)
#define insertfinish (0)

#ifdef LINES
int lines = LINES, lines_1 = LINES - 1;
#else
int lines = 24, lines_1 = 23;
#endif
int cols = 80, cols_1 = 79;

#endif /* !VT100 */

int col0;		/* input line offset (= length of prompt) */
int line0;		/* screen line where the input line starts */

#ifdef USETERMIO
static termiostruct ttybsave;
#else
static struct sgttyb ttybsave;
static struct tchars tcsave;
static struct ltchars ltcsave;
#endif

/*
 * Terminal handling routines:
 * These are one big mess of left-justified chicken scratches.
 */

/*
 * Save the original state, disable echo and special keys.
 */
void set_terminal()
{
#ifdef USETERMIO
    ioctl(0, TCGETS, &ttybsave);
#else /* not USETERMIO */
    ioctl(0, TIOCGETP, &ttybsave);
    ioctl(0, TIOCGETC, &tcsave);
    ioctl(0, TIOCGLTC, &ltcsave);
#endif
    set_spec_keys(0);
    fputs(kpadstart, stdout);
}

/*
 * Reset the terminal to its original state
 */
void reset_terminal()
{
#ifdef USETERMIO
    ioctl(0, TCSETS, &ttybsave);
#else
    ioctl(0, TIOCSETP, &ttybsave);
    ioctl(0, TIOCSETC, &tcsave);
    ioctl(0, TIOCSLTC, &ltcsave);
#endif
    fputs(kpadend, stdout);
}

/*
 * set raw/cooked mode depending on the current linemode
 * enable/disable special keys
 */
void set_spec_keys(k)
int k;
{
#ifdef USETERMIO
    termiostruct ttyb;
    ttyb = ttybsave;
    ttyb.c_cc[VTIME] = 0;
    ttyb.c_cc[VMIN] = 1;
    ttyb.c_oflag |= OPOST;
    if(linemode & LM_CHAR) ttyb.c_oflag &= ~OPOST;
    ttyb.c_lflag &= ~(ECHO | ISIG | ICANON | IEXTEN);
    if(k) ttyb.c_lflag |= ISIG;
    ioctl(0, TCSETS, &ttyb);
#else /* not USETERMIO */
    static struct tchars tc_off = { -1, -1, -1, -1, -1, -1 };
    static struct ltchars ltc_off = { -1, -1, -1, -1, -1, -1 };
    struct tchars *tc = &tc_off;
    struct ltchars *ltc = &ltc_off;
    struct sgttyb ttyb;
    ttyb = ttybsave;
    ttyb.sg_flags &= ~(O_RAW | O_ECHO);
    if(linemode & LM_CHAR) ttyb.sg_flags |= O_RAW;
    if(k) {
       tc = &tcsave;  ltc = &ltcsave;
    }
    ioctl(0, TIOCSETP, &ttyb);
    ioctl(0, TIOCSETC, tc);
    ioctl(0, TIOCSLTC, ltc);
#endif
}

/*
 * read termcap definitions
 */
void settermcap()
{
#ifndef VT100
    struct tc_init_node {
	char cap[3], critic;
	char *buf;
	int *len;
    };
    static struct tc_init_node tc_init[] = {
	{ "cm", 1, curgoto, 0 },
	{ "ce", 1, clreoln, &len_clreoln },
	{ "cd", 1, clreoscr, 0 },
	{ "nd", 1, rightcur, 0 },
	{ "le", 0, leftcur, &len_leftcur },
	{ "up", 0, upcur, &len_upcur },
	{ "cr", 0, begoln, &len_begoln },
	{ "im", 0, inschar, &insertcost },
	{ "ic", 0, inschar, &insertcost },
	{ "ei", 0, insstop, &insertcost },
	{ "dm", 0, delchar, &deletecost },
	{ "dc", 0, delchar, &deletecost },
	{ "ed", 0, delchar, &deletecost },
	{ "me", 0, modenorm, 0 },
	{ "md", 0, modebold, 0 },
	{ "mr", 0, modeinv, 0 },
	{ "us", 0, modeuline, 0 },
	{ "so", 0, modestandon, 0 },
	{ "se", 0, modestandoff, 0 },
	{ "ks", 0, kpadstart, 0 },
	{ "ke", 0, kpadend, 0 },
	{ "kl", 0, key_left, 0 },
	{ "kr", 0, key_right, 0 },
	{ "ku", 0, key_up, 0 },
	{ "kd", 0, key_down, 0 },
	{ "",   0, NULL, 0 }
    };
    struct tc_init_node *np;
    char tcbuf[2048];		/* by convention, this is enough */
    int i;
#endif    
#if !defined(VT100) || defined(TELNETBUG)
    char *term = getenv("TERM");
    if(!term) {
	printf("$TERM not set\n"); 
	exit(1);
    }
#endif
#ifndef VT100
    switch(tgetent(tcbuf, term)) {
    default:
	syserr("tgetent");
    case 0:
	fprintf(stderr,
	    "There is no entry for '%s' in the terminal data base.\n", term);
	fprintf(stderr,
	    "Please set your $TERM environment variable correctly.\n");
	exit(1);
    case 1:
	break;
    }
    for(np = tc_init; np->cap[0]; np++)
	if((i = extract(np->cap, np->buf))) {
	    if(np->len) *np->len += i;
	} else if(np->critic) {
	    fprintf(stderr,
		    "Your terminal is not powerful enough, missing '%s'.\n",
		    np->cap);
	    exit(1);
	}
    if(!len_begoln)
	strcpy(begoln, "\r"), len_begoln = 1;
    if(!len_leftcur)
	strcpy(leftcur, "\b"), len_leftcur = 1;
    cols_1 = (cols = tgetnum("co")) - 1;
    lines_1 = (lines = tgetnum("li")) - 1;
    wrapglitch = tgetflag("xn");
    metakey = tgetflag("km");
    gotocost = strlen(tgoto(curgoto, cols_1, lines_1));
    insertfinish = gotocost + len_clreoln;
#endif
#ifdef TELNETBUG
    if(strncmp(term, "vt10", 4) == 0) {
	/* might be NCSA Telnet 2.2 for PC, which doesn't reset colours */
	sprintf(modenorm, "\033[;3%d;4%dm", DEFAULTFG, DEFAULTBG);
    }
#endif
}

/*
 * add the default keypad bindings to the list
 */
void keypad_bindings()
{
    /*
     * Note: termcap doesn't have sequences for the numeric keypad, so we just
     * assume they are the same as for a vt100.
     */
    add_keynode("kp2", "\033Or", key_run_command, "s");
    add_keynode("kp3", "\033Os", key_run_command, "d");
    add_keynode("kp4", "\033Ot", key_run_command, "w");
    add_keynode("kp5", "\033Ou", key_run_command, "exits");
    add_keynode("kp6", "\033Ov", key_run_command, "e");
    add_keynode("kp8", "\033Ox", key_run_command, "n");
    add_keynode("kp9", "\033Oy", key_run_command, "u");
}
    
/*
 * initialize the key binding list
 */
void default_bindings()
{
    struct defbind_node *p = defbinds;
    /* FEATURE: no check is made to see if (up, down, etc.) keys have valid
      keysequence. could use key_collision(), which is defined in cmd.c */
    do {
	if(p->seq[0]) add_keynode(p->label, p->seq, p->funct, NULL);
    } while((++p)->label[0]);
}

#ifndef VT100
/*
 * extract termcap 'cap' and strcat it to buf.
 * return the lenght of the extracted string.
 */
static int extract(cap, buf)
char *cap, *buf;
{
    static char *bp;
    char *d = buf + strlen(buf);
    char *s = tgetstr(cap, (bp = d, &bp));
    int len;
    if(!s) return (*bp = 0);
    /*
     * Remove the padding information. We assume that no terminals
     * need padding nowadays. At least it makes things much easier.
     */
    s += strspn(s, "0123456789*");
    for(len = 0; *s; *d++ = *s++, len++)
	if(*s == '$' && *(s + 1) == '<')
	    if(!(s = strchr(s, '>')) || !*++s) break;
    *d = 0;
    return len;
}
#endif /* VT100 */

/*
 * position the cursor using absolute coordinates
 */
void gotoxy(col, line)
int col, line;
{
#ifdef VT100
  printf("\033[%d;%dH", line + 1, col + 1);
#else
  fputs(tgoto(curgoto, col, line), stdout);
#endif
}

/*
 * optimized cursor movement
 * from (fromline, fromcol) to (toline, tocol)
 * if tocol > 0, (toline, tocol) must reside on editline.
 */
void optmove(fromline, fromcol, toline, tocol)
int fromline, fromcol, toline, tocol;
{
    static char buf[CAPLEN];
    char *cp = buf;
    int i, dist, cost = 0;

    /* First, move vertically to the correct line, then horisontally
     * to the right column. If this turns out to be fewer characters
     * than a direct cursor positioning (gotoxy), use that.
     */
    for(;;) {	/* gotoless */
	if((i = toline - fromline) < 0) {
	    if(!len_upcur || (cost = -i * len_upcur) >= gotocost) break;
	    do strcpy(cp, upcur), cp += len_upcur; while(++i);
	} else if(i) {
	    /* all lf's mapped into crlf on output */
	    if((cost = 2 * i) >= gotocost) break;
	    do *cp++ = '\n'; while(--i);
	    fromcol = 0;
	}
	if((i = tocol - fromcol) < 0) {
	    dist = -i * len_leftcur;
	    if(dist <= len_begoln + tocol) {
		if((cost += dist) > gotocost) break;
		do strcpy(cp, leftcur), cp += len_leftcur; while(++i);
	    } else {
		if((cost += len_begoln) > gotocost) break;
		strcpy(cp, begoln);  cp += len_begoln;
		fromcol = 0;  i = tocol;
	    }
	}
	if(i) {
	    /* if hiliting in effect, just use gotoxy */
	    if(cost + i > gotocost || *edattrbeg) break;
	    if(fromcol < col0 && toline == line0) {
		/* FEATURE: it should be recorded somewhere whether printed
		  prompt was marked. In that case, gotoxy must be used */
		if(promptstr[col0]) break;	/* special chars in prompt */
		strcpy(cp, promptstr);
		cp += strlen(cp);
		fromcol = col0;
	    }
	    strncpy(cp, edbuf + (toline - line0) * cols_1 + fromcol - col0,
		    tocol - fromcol);
	    cp += tocol - fromcol;
	}
	*cp = 0;
	fputs(buf, stdout);
	return;
    }
    gotoxy(tocol, toline);
}

/*
 * change the position on input line (optmove there, and set pos)
 */
void moveto(new_pos)
int new_pos;
{
    /*
     * FEATURE: the line we are moving to might be less than 0, or greater
     * than lines_1, if the display is too small to hold the whole editline.
     * In that case, the input line should be (partially) redrawn.
     */
    optmove(CURLINE(pos), CURCOL(pos), CURLINE(new_pos), CURCOL(new_pos));
    pos = new_pos;
}

/*
 * move to the beginning of last line
 */
void lastline()
{
    /* NOTE: could use tcap "ll" here */
    gotoxy(0, line0 = lines_1);
}

/*
 * display a one-line footnote at the bottom of the screen
 */
void show_footnote()
{
    int to_c = CURCOL(pos);
    int to_l = CURLINE(pos);
    if(edattr) { edattr = 0; fputs(edattrend, stdout); }
    optmove(to_l, to_c, lines_1, 0);
    if(CURLINE(edlen) >= lines_1) {
	line0--;  to_l--;  putchar('\n');
    }
    fputs(footnote, stdout);
    gotoxy(to_c, to_l);
}

/*
 * clear the footnote or put cursor back where it belongs
 */
void clear_footnote()
{
    int to_c = CURCOL(pos);
    int to_l = CURLINE(pos);
    optmove(to_l, to_c, lines_1, 0);
    fputs(clreoln, stdout);
    optmove(lines_1, 0, to_l, to_c);
    *footnote = 0;
}

/*
 * clear current input line (but not the prompt)
 * cursor is left right after the prompt
 */
void clearinpline()
{
    if(edlen) {
	int from_c = CURCOL(pos);
	int from_l = CURLINE(pos);
	if(line0 < lines_1) {
	    optmove(from_l, from_c, line0 + 1, 0);
	    fputs(clreoscr, stdout);
	    from_c = 0;  from_l = line0 + 1;
	}
	optmove(from_l, from_c, line0, col0);
	fputs(clreoln, stdout);
    }
}

/*
 * Redraw the input line and put the cursor at the current position.
 * The cursor is assumed to be directly after the prompt.
 */
void redraw_input()
{
    if(edlen) {
	int n, k = cols_1 - col0;
	if(!edattr && *edattrbeg) {
	    edattr = 1;  fputs(edattrbeg, stdout);
	}
	printf("%.*s", k, edbuf);
	if(edlen >= k) {
	    n = lines_1;
	    do {
		printf("\n%.*s", cols_1, edbuf + k);
		--n;  k += cols_1;
	    } while(k <= edlen);
	    if(n < line0) line0 = n;
	}
	n = pos;  pos = edlen;
	moveto(n);
    }
}

/*
 * delete n characters at current position (the position is unchanged)
 * assert(n <= edlen - pos)
 */
void deletechars(n)
int n;
{
    int r_cost, d_cost = n, p = pos;
    int nl = p - CURCOL(p);	/* next line's starting pos */
    int cl = CURLINE(p);	/* current line */
    for(d_cost = p + n; d_cost < edlen; d_cost++)
	edbuf[d_cost - n] = edbuf[d_cost];
    for(d_cost -= n; d_cost < edlen; d_cost++)
	edbuf[d_cost] = ' ';
    for( ; ; putchar('\n'), p = nl, cl++) {
	d_cost = 0;
	/* NOTE: ought to be "d_cost = n > gotocost ? -gotocost : -n;",
	  since redraw will need to goto back. Of little importance */
	if((r_cost = edlen) > (nl += cols_1))
	    r_cost = nl, d_cost = n + gotocost;
	r_cost -= p;
#ifndef VT100
	/* NOTE: no clreoln is used (it might cost less in the
	  occasion we delete more than one char). Simplicity */
	if(deletecost && deletecost * n + d_cost < r_cost) {
	    for(d_cost = n; d_cost; d_cost--) fputs(delchar, stdout);
	    if(edlen <= nl) break;
	    gotoxy(cols_1 - n, cl);
	    printf("%.*s", n, edbuf + nl - n - 1);
	} else
#endif
	{
	    printf("%.*s", r_cost, edbuf + p);  p += r_cost;
	    if(edlen <= nl) break;
	}
    }
    edbuf[edlen -= n] = '\0';
    switch(pos - p) {
	default: optmove(cl, CURCOL(p), CURLINE(pos), CURCOL(pos)); break;
	case 1: fputs(leftcur, stdout); break;
	case 0: break;
    }
}

/*
 * insert n characters at current position. The position is set to after
 * the inserted characters.
 */
void insertchars(str, n)
char *str; int n;
{
    int r_cost, i_cost = n, p = pos;
    int nl = p - CURCOL(p);	/* next line's starting pos */
    int cl = CURLINE(p);	/* current line */
    while(i_cost) strins(edbuf + p, str[--i_cost]);
    edlen += n;  pos += n;
    do {
	i_cost = n;
	if((r_cost = edlen) > (nl += cols_1))
	    r_cost = nl, i_cost += insertfinish;
	r_cost -= p;
#ifndef VT100
	/* NOTE: insert mode is used only when one char is inserted
	  (which is probaly true > 95% of the time). Simplicity */
	if(n == 1 && insertcost && insertcost + i_cost < r_cost) {
	    printf("%s%c%s", inschar, edbuf[p++], insstop);
	    if(edlen > nl) {
		gotoxy(cols_1, cl);
		fputs(clreoln, stdout);
	    }
	} else
#endif
	{
	    printf("%.*s", r_cost, edbuf + p);  p += r_cost;
	}
	if(edlen < nl) break;
	putchar('\n');  p = nl;
	if(++cl > lines_1) cl = lines_1, line0--;
    } while(edlen > nl);
    if(p != pos) optmove(cl, CURCOL(p), CURLINE(pos), CURCOL(pos));
}

/*
 * insert one character first in a string
 */
static void strins(s, c)
char *s; char c;
{
  char d;
  for(; *s; d = *s, *s++ = c, c = d);
  *s++ = c; *s = 0;
}

/*
 * print a newline and advance line0.
 */
void newline()
{
    putchar('\n');
    if(line0 < lines_1) line0++;
}

/*
 * search for marker in string
 * marker must not be empty string
 *
 * FEATURE: str_mark ought perhaps match a marker starting/ending with a
 * word separator char regardless if it's preceded/followed by one or not.
 */
static char *str_mark(str, mark)
char *str, *mark;
{
    register char *s = str;
    register char *m = mark;
    if(*s) for( ; ; ) {
	if(*m++ == *s++)
	    if(*m) continue;
	    else if(!islabel(*s)) return str;
	m = mark;  s = str;
	while(*++s && (*s != *m || islabel(*(s - 1))));
	if(!*(str = s)) break;
    }
    return NULL;
}

/*
 * put escape sequences to turn on/off an attribute in given buffers
 */
void attr_string(attrcode, begin, end)
int attrcode; char *begin, *end;
{	
    int fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode);
    *begin = *end = '\0';
    if(fg < COLOURS) {
	if(bg < COLOURS)
	    sprintf(begin, "\033[3%d;4%dm", fg, bg);
	else
	    sprintf(begin, "\033[3%dm", fg);
    } else if(bg < COLOURS) {
	sprintf(begin, "\033[4%dm", bg);
    } else {
	char *src;
	if(fg == UNDERLINE) src = modeuline;
	else if(fg == BOLD) src = modebold;
	else if(fg == INVERSE) src = modeinv;
	else return;
	if(!*src) {
	    strcpy(begin, modestandon);
	    strcpy(end, modestandoff);
	    return;
	}
	strcpy(begin, src);
    }
    if(attrcode != NOATTRCODE)
	strcpy(end, modenorm);
}

/*
 * print line, scanning for marked text
 *
 * FEATURE: mark_and_print and wrap_out ought to be integrated. Right now,
 * text which is otherwise marked, is not marked when it's wrapped.
 * Marking is performed regardless of possible attribute changes in the text.
 * Cancan uses attributes, but doesn't really handle them. For instance, try
 * "#mark \ =bold", and see what output is produced for long space sequences.
 *
 * BUG: It is possible that a letter ending an escape sequence is matched as
 * the first char of a marker.
 */
void mark_and_print(line)
char *line;
{
    static char begin[CAPLEN], end[CAPLEN];
    marknode *m, *np;

    for(m = 0, np = markers; np; np = np->next)
	if((np->pos = str_mark(line, np->pattern)))
	    if(!m || m->pos > np->pos) m = np;
    /* NOTE: ^ no need to match the rest of the markers if m->pos == line ! */
    while(m) {
	attr_string(m->attrcode, begin, end);
	printf("%.*s%s%s%s", m->pos - line, line, begin, m->pattern, end);
	if(!*(line = m->pos + strlen(m->pattern))) break;
	for(m = 0, np = markers; np; np = np->next)
	    if(np->pos)
		if(np->pos >= line || (np->pos = str_mark(line, np->pattern)))
		    if(!m || m->pos > np->pos) m = np;
    }
    fputs(line, stdout);
}

/*
 * write string to stdout, wrapping words to the next line.
 * attempt to handle the control characters.
 * return the position in the string where (last) wrap occurred, or NULL.
 *
 * FEATURE: ansi-like escape sequences assumed -- not entirely universal.
 * BUG: incomplete escape sequence at the end of the string is printed...
 */
char *wrap_out(str)
char *str;
{
    char *tail = 0;
    if(*str) {
	unsigned int width = wrap ? cols_1 + wrapglitch : (unsigned)-1;
	unsigned int cc = col0;
	char *s = str, *wp, c;
	do {
	    for(wp = 0, s--; ; ) {
		if(isprint(*++s)) {
		    if(*s == ' ') wp = s;
		    if(cc++ < width) continue;
		} else switch(*s) {
		    default: continue;
		    case '\033':if(*++s == '[') while(*++s && !isalpha(*s));
				if(*s) continue;
		    case '\0':	break;
		    case '\b':	if(cc-- > 1) continue;
		    case '\r':	cc = 0;  tail = s;  continue;
		    case '\t':	wp = s;  cc |= 7;
				if(++cc < width) continue;
		}
		break;
	    }
	    if(!*s) { mark_and_print(str);  break; }
	    if(wp) s = wp;
	    c = *s;  *s = '\0';
	    mark_and_print(str);  *s = c;
	    if(wp) do s++; while(*s == ' ');
	    cc = 0;  newline();
	    tail = str = s;
	} while(*s);
	col0 = cc;
    }
    return tail;
}

/*
 * window size change signal handler:
 * we assume that the bottom line is always the bottom line
 */
void winchsig()
{
    struct winsize wsiz;
  
    /* if ioctl fails or gives silly values, don't change anything */
    if(ioctl(0, TIOCGWINSZ, &wsiz) == 0 && wsiz.ws_row && wsiz.ws_col > 1) {
	lines = wsiz.ws_row; lines_1 = lines - 1;
	cols = wsiz.ws_col; cols_1 = cols - 1;
	/* NOTE: to which session are we sending it?? */
	send_term_size();
    }
    /* reinstall signal handler for system V unices */
    signal(SIGWINCH, winchsig);
}
